# $Id: ConfigFile.pm,v 3.7 1998-06-01 14:46:34-05 annis Exp $

=head1 NAME

ConfigFile - reads a configuration file and dumps its information into
a table.

This is RCS version: $Revision: 3.7 $.

=head1 DESCRIPTION

C<ConfigFile::new> creates a single object which opens a specified
configuration file and places all this information into tables which
can then be accessed through simple methods.  The point of this
interface is to make it easy to create useful configuration files for
differing system adminstration tasks that need to be accomplished on
numerous machines.

=head1 FORMAT

The configuration file format is generalized into two sections:
C<DEFAULT> blocks which specify default behaviors (I<i.e.> default
properties) which may be included by C<NAME> blocks which specify the
default behaviors for specific entities (machines, for example).
Comments are as in shell and Perl: anything after a #.

The C<DEFAULT> block looks like this:

    DEFAULT default_tag
        def_property0 def_value0
        def_property0 def_value0
        ...
        def_propertyN def_valueN
    END

The I<default_tag> may be any atomic string and is used by C<NAME>
blocks to include properties.

A C<NAME> block looks like this:

    NAME machine_name
        INCLUDE some_tag
        INCLUDE some_other_tag
        property0 value0
        property1 value1
        ...
        propertyN valueN
    END

B<NOTE:> it is important to know that when asking for the value of a
property for an entry, the explicit properties are checked first, then
the include blocks are checked I<in the order they are specified in the block>.
This is to allow you to override defaults on a particular
entry without having to reenter other useful defaults

Information is extracted from the file via the C<get> method.  From
the above example you could call

    $answer = $config->get('machine_name', 'property0')

and you would get C<'value0'> back as your answer.

If the flexibility (and associated syntax) of the C<NAME> blocks is
too much for your application, you may instead use the C<GLOBAL>
directive.  For example:

    GLOBAL

    port        5001
    other-port  5002
    tmp         /tmp/myapp

The C<GLOBAL> directive must be at the end of the configuration file.
The C<get> syntax for C<GLOBAL>s is simpler:

    $answer = $config->get('port');

The values of properties may be extended over multiple lines if in the
C<value> position you place the token '@{'.  All following lines,
until a matching '@}' is found are assigned to the propery.  For
example:

    NAME button-1
      INCLUDE button-defaults
      Command @{
          $my_name = &some_perl_code($id, $v1);
          if (! defined $my_name) {
              exit(1);
          } else {
              &wahoo($my_name);
          }
      @}
      Label  Quit
    END

Text included between '@{' and '@}' does NOT get tokenized, comments
are ignored (that is, included in the property), but macro
substitution does take place.

=head1 MACROS

In addition to the simple property/value association assignment, this
configuration file tool allows basic macro substitution and
conditional assignment of values to properties.

B<NOTE:> all the @@-directives must be syntactically complete on a
single line.  That is, you cannot spread a complex conditional or
assignment statement over several lines.

A value is assigned to a macro in the following way:

    @@define symbol This is a value.

The C<@@define> directive has global scope, it can occur at the top
level or in a C<DEFAULT> or C<NAME> block, but is in effect for all
sections of the configuration file following it.

Macros may be accessed by prefixing the symbol name with C<@$>, and
may occur anywhere in the file.  So C<@$symbol> will have the string
C<"This is a value."> substituted for it.

In addition to macros defined with the C<@@define> syntax, it is now
possible to use a property previously defined in the current block as
a macro.  For example:

    NAME system
        pk	hostname
        sql	select @$pk from system;
    END

In the above defintion, the C<sql> property will expand to 'select
hostname from system;'.  Also, the name of the name block ('system'
above) is available as the macro C<@$NAME> which is thus a reserved
macro.  So, you could write the above block as following with the
same results.

    NAME system
        pk	hostname
        sql	select @$pk from @$NAME;
    END

=head2 Macro Conditionals

There is a primitive conditional facility which is intended to
interact with the macros:

    @@ifdefined  SOME_VAR ...
    @@ifndefined SOME_VAR ...
    @@ifequal    val1 val2  ...
    @@ifnequal   val1 val2  ...

Where...

    val ::= { @$VARIABLE | atomic-string }
    ... ::= { @@directive | property value }

Notice that you can do things like the following:

    @@ifequal @$OS_SHORT solaris   @@define PS_FLAGS -ef

or this:

    NAME bob
      @@ifequal @$OS_SHORT solaris   ps_flags -ef
      @@ifequal @$OS_SHORT sunos     ps_flags -auxw
    END

This last form assigns the value C<-ef> to the C<NAME bob> block
property C<ps_flags> if the program is run on a Solaris machine.  It
is an error to use this form outside a block.

Notice the syntax for C<@@ifdefined>:

    @@ifndefined HOSTNAME   @@define BRAIN_DAMAGE hostname-damage

You do B<not> prefix the macro name with C<@$> when testing if it is
defined or not.

B<EXPERIMENTAL:> Certain macros are defined when the ConfigFile object
is created.  As of this version, only SunOS and Solaris will assign
all of these properly.

    @$COMMENT		- The character '#'.
    @$DATE		- The date/time when the config file was read.
    @$DATETAG		- YYMonDD
    @$OS		- 'uname'
    @$HOSTNAME		- 'uname -n'
    @$OS_VERSION	- 'uname -r'
    @$OSVERSION		- @$OS . @$OS_VERSION (i.e. SunOS5.5.1)
    @$ARCH		- 'uname -p'
    @$OS_SHORT		- sunos | solaris

Of course, these may be overwritten in your own configuration files.

=head1 VARIABLES

=head2 C<$ConfigFile::ERROR>

This variable is used by C<new> and C<read> when there are problems
with the configuration file specified.

=head1 METHODS

=head2 C<new>

    $my_config = new ConfigFile(File => 'config.txt');
    $my_config = new ConfigFile;

This opens and parses the configuration file specified.  Will return
C<undef> on errors, with the error string in C<$ConfigFile::ERROR>.

=head2 C<read>

    $my_config->read('config.txt');

This reads the specified config file.  All the interesting work
happens here.  One useful trick you can use this for is to read in
several configuration files, one of which may simply assign macro
values.

Will return C<undef> on errors, with the error string in
C<$ConfigFile::ERROR>.


=head2 C<set_default>

    $my_config->set_default('default_tag');

If you call C<get> for an entity which does not have an entry the
default return value is C<undef>.  This call sets up a default set of
properties for entities without explicit C<NAME> blocks.

=head2 C<get>

    $value = $my_config->get('name', 'property');
    $value = $my_config->get('name', 'property', 'default_tag');
    $value = $my_config->get('name');

The first form of C<get> looks for the value and returns C<undef> if
there is no such entry or property.  The second form looks for the
property of a entity, but if no such entity is specified in a C<NAME>
block, returns the value for the default block, but will return
C<undef> if there is no such C<DEFAULT default_tag> block or property.
See above for in C<set_default> for a nicer way to deal with defaults.

The last form is for GLOBAL properties.

=head2 C<get_all>

    @list = $my_config->get_all('name');
    @list = $my_config->get_all;

This returns a list of all non-default properties for the specified
block.  The second form, without arguments, returns a list of all
GLOBAL properties.

Note that this returns a list of the property names, not the actual
values.

=head2 C<set>

    $my_config->get('name', 'property', $value);

This assigns the value of a property for a specific entry.

=head2 C<add_default>

    $my_config->add_default('name', 'default_tag');

This acts exactly like a C<DEFAULT some_tag> clause in a C<NAME>
block.  This is the way to add finer grained default behavior than is
offered by C<set_default>.

=head2 C<defaults>

    @defaults = $my_config->defaults('name');

Returns a list of DEFAULTs available to this NAME block.  Returns
C<undef> if there is no such NAME or if it has no DEFAULTs.

=head2 C<names>

    @names = $my_config->names();

Returns a list of all NAME blocks defined.  Note that this will not
tell you if a C<GLOBAL> block is in use.

=head2 C<define>

    $my_config->define('macro', 'value');

This corresponds to the C<@@define> directive in the configuration
language.  Obviously, this is only useful before a call to C<read>.
For the above example, you will need to reference "C<@$macro>" to get
interpolation.

=head1 AUTHOR

William S. Annis <annis@biostat.wisc.edu>.

L<http://www.biostat.wisc.edu/~annis>

Copyright (c) 1997, 1998 William S. Annis.  All rights reserved.  This
is free software; you can redistribute it and/or modify it under the
same terms as Perl itself (the Artistic Licence).

=head1 HISTORY

A large part of this code was extracted from a program I wrote to
monitor the disk use on an army of machines.  I found the
configuration file parsing and tabling took up a very large part of
that code.

This module has subsequently been used in a number of applications.
It is used in all parts of my machine monotoring system, Mom.  It has
also been put to quite peculiar use as a definition language for GUIs
using PerlTk.

=head1 COMMENTS and BUGS

The parsing of the file could be a little less delicate and cranky.
In particular, tokens must be separated by whitespace.  Bad syntax
with multi-line properties returns errors with line numbers rather
distant from the real problem.

This documentation could be clearer.

=cut

# ConfigFile.pm - a class for configuration files and the tables who love them.
# Author: William Annis, Biomedical Computing Group, annis@biostat.wisc.edu
# Started: Mon Mar 24 14:06:39 1997

package ConfigFile;
use vars qw($VERSION $ERROR);
$VERSION = 1.3;

# This is used during file parsing.
my $line_number;

# Contains error strings from Die.  This way, 'new' can return undef
# on failure.
$ERROR = "";

# For error messages:
my $config_file;

# Pseudo-NAME block name for global properties.
my $GLOBAL_NAME_ID = " GLOBAL ";

# This keeps track of all variable/macros defined.
my %variables;

sub new {
    my($class, %args) = @_;
    my $self = {};
    my $file = $args{'File'};

    # Make ourselves a class.
    bless $self, $class;

    # Setup the global default DEFAULT.
    undef $self->{'default DEFAULT'};

    # Set up default variable names.
    &define_default_variables;

    # Go look at the file.
    if (defined $file) {
        my $ok = &read($self, $file);
        return undef if ! $ok;
    }

    return $self;
}


# Backward hack.  This allows you to create a ConfigFile object
# without using the ugly "File =>" syntax.
sub read {
    my ($self, $file) = @_;

    if ((! defined $file ) || (! -r $file)) {
        $ERROR = "Cannot read: $file.  Doesn't exist or is unreadable.";
        return undef;
    } else {
        # Assign this for use by error messages:
        $config_file = $file;

        # Catch dies from the parser and return undef.  $ERROR will be
        # assigned from the $@ value (string to die).
        eval {
            &ponder_config_file($self, $file);
        };
        if ($@) {
            $ERROR = $@;
            return undef;
        } else {
            # Must be explicit here... everything is fine.
            return 1;
        }
    }
}

# Get the property set for a machine.
sub get {
    my ($self, $name, $property, @rest) = @_;

    # If there is only one argument ($name), then the user is trying
    # to get at a global property.
    if (! defined $property) {
        $property = $name;
        $name = $GLOBAL_NAME_ID;
    }
    my $value = $self->{$name}{$property};

    # If we didn't find a value, start looking.
    if (! defined $value) {
        # First, check the built-in default clauses.
        if (defined $self->{$name}{'DEFAULT'}) {
            my (@defaults) = split / /,$self->{$name}{'DEFAULT'};
            if ($#defaults != -1) {
                foreach $def (@defaults) {
                    $value = $self->{"DEFAULT $def"}{$property};
                    if (defined $value) {
                        return $value;
                    }
                }
            }
        }

        # Is there a default DEFAULT?
        my $def_default = $self->{'default DEFAULT'};
        if (defined $def_default) {
            $value = $self->{"DEFAULT $def_default"}{$property};
        }

        # Well, we didn't find it there.  Try the last argument?
        if ((! defined $value) && ($#rest != -1)) {
            $value = $self->{"DEFAULT $rest[0]"}{$property};
        }
    }
    
    # Well, return whatever we have.
    return $value;
}


# Return a list of names of all explicitly defined properties.
sub get_all {
    my ($self, $name) = @_;

    if (! defined $name) {
        $name = $GLOBAL_NAME_ID;
    }

    # Get copy of hash.
    my %value = %{$self->{$name}};

    # Remove housekeeping properties.
    delete $value{'DEFAULT'};
    delete $value{'ConfigFile::entry'};

    keys %value;
}

# Set the global default DEFAULT for all entities.
sub set_default {
    my ($self, $default_tag) = @_;
    
    $self->{'default DEFAULT'} = $default_tag;
}

# Change or add an entry to an entity.
sub set {
    my ($self, $id, $prop, @value) = @_;

    $self->{$id}{$prop} = "@value";
}

# Assign a DEFAULT tag to an entity.
sub add_default {
    my ($self, $id, $default_tag) = @_;
    
    $self->{$id}{'DEFAULT'} = $self->{$id}{'DEFAULT'} . " $default_tag";
}

# Get a list of all the defaults.
sub defaults {
    my ($self, $name) = @_;

    my $value = $self->{$name}{'DEFAULT'};
    if (defined $value) {
        split / /, $value;
    } else {
        undef;
    }
}

# Returns a list of all NAME blocks.
sub names {
    my $self = shift;

    my @ret;
    foreach my $name (keys %{$self}) {
        next if $name =~ /^DEFAULT /;
        next if $name eq $GLOBAL_NAME_ID;
        next if $name eq "default DEFAULT";

        push @ret, $name;
    }
    
    @ret;
}

# Define a macro for use by subsequent reads.  The $table argument, if
# given, must be a hash reference, into which the macro definitions
# will go (see &config_tokenize for the use of this).
sub define {
    my ($self, $macro, $val, $table) = @_;

    if (! defined $table) {
        $variables{"\\\@\\\$$macro"} = $val;
    } else {
        $table->{"\\\@\\\$$macro"} = $val;
    }
}

# Unadvertised function: parse the config file and shove stuff into the
# property tables.  All the fun is from here down.
sub ponder_config_file {
    my ($self, $file) = @_;

    $line_number = 0;

    open(CONFIG, "< $file") or die "Cannot open $file";
    while(<CONFIG>) {
        $line_number++;

        next if /^$/;           # Skip a blank line.
        next if /^\s*#/;        # Skip a comment line
        
        my @tokens = &config_tokenize($_);
        next if ($#tokens == -1); # Skip lines of spaces and tabs.

        if ($tokens[0] =~ /^\@\@/) {
            # Read an @@-directive.
            &config_handle_directive($self, \*CONFIG, undef, @tokens);
            next;
        }
        elsif ($tokens[0] eq "DEFAULT") {
            # Do DEFAULT block.
            &config_read_default($self, $tokens[1], \*CONFIG);
        }
        elsif ($tokens[0] eq "NAME") {
            # Do MACHINE block.
            &config_read_name($self, $tokens[1], \*CONFIG);
        } elsif ($tokens[0] eq 'GLOBAL') {
            # Read GLOBAL properties.
            &config_read_name($self, $GLOBAL_NAME_ID, \*CONFIG);
        } else {
            die "Malformed configuration file $config_file.\nNear line $line_number : @tokens\nExpecting DEFAULT or NAME.\n"; 
        }
    }

    close CONFIG;
}

# Get the line and return as tokens, tossing away any comments.
sub config_tokenize {
    my ($line, $locals) = @_;
    
    # Look at the line to make the regexps faster.
    study($line);

    # Deal with comments.
    $line =~ s/#.*//g;

    # Substitute @$-variables in here.
    my $key;

    # If there are locals, use them, too.
    if (defined $locals) {
        foreach $key (keys %{$locals}) {
            $line =~ s/$key/$locals->{$key}/g;
        }
    }

    # Now, the @@define'd macros.
    foreach $key (keys %variables) {
        $line =~ s/$key/$variables{$key}/g;
    }

    # Remove leading spaces and tabs.
    $line =~ s/^\s*//g;

    # Remove excess spaces between fields (they confuse split).
    $line =~ s/\s\s*/ /g;

    # Split it up and return.
    return split(/\s/, $line);
}

# Read and install data for a DEFAULT block.  Takes (self, default_tag,
# filehandle).
sub config_read_default {
    my $self   = shift;
    my $tag    = shift;
    my $CONFIG = scalar shift;

    # Local macro substitution.
    my %locals;
    $self->define('NAME', $tag, \%locals);

    while (<$CONFIG>) {
        $line_number++;
        next if /^$/;           # Skip a blank line.
        next if /^\s*#/;        # Skip a comment line

        @tokens = &config_tokenize($_, \%locals);
        next if ($#tokens == -1);
        
        if ($tokens[0] =~ /^\@\@/) {
            # Read an @@-directive.
            &config_handle_directive($self, \*CONFIG, "DEFAULT $tag", @tokens);
            next;
        }
        # One more validation: no DEFAULT clauses allowed in a DEFAULT block.
        if ($tokens[0] eq 'DEFAULT' || $tokens[0] eq 'INCLUDE') {
            die "Line $line_number invalid: 'INCLUDE' not allowed in 'DEFAULT' block.\n";
        }

        # Pound the information into the property table.
        if (($#tokens == 0) && ($tokens[0] eq 'END')) {
            # Should be the 'END'.
            return;
        } elsif ($#tokens >= 1) {
            my $property = shift @tokens;
            if ($tokens[0] eq '@{') {
                my @prop = &read_property_block(\*CONFIG, \%locals);
                $self->{"DEFAULT $tag"}{$property} = "@prop";
            } else {
                $self->{"DEFAULT $tag"}{$property} = "@tokens";
            }

            # Shove the property just found into the local macro table.
            $self->define($property,  $self->{"DEFAULT $tag"}{$property}, \%locals);
        } else {
            die "Malformed configuration file near line $line_number:\n\t@tokens\n\tExpecting 'END'?";
        }
    }
}

# Read and install data for a NAME clause.  Takes (self, name, filehandle).
sub config_read_name {
    my $self = shift;
    my $name = shift;
    my $CONFIG = scalar shift;

    # First, say that the machine has an entry.
    $self->{$name}{'ConfigFile::entry'} = 1;

    # Namespace for local macros.
    my %locals;
    $self->define('NAME', $name, \%locals);

    while (<CONFIG>) {
        $line_number++;
        next if /^$/;           # Skip a blank line.
        next if /^\s*#/;        # Skip a comment line

        @tokens = &config_tokenize($_, \%locals);
        next if ($#tokens == -1);

        if ($tokens[0] =~ /^\@\@/) {
            # Read an @@-directive.
            &config_handle_directive($self, \*CONFIG, "NAME $name",  @tokens);
            next;
        }
        # Shove this info into the property tables.
        if ($#tokens == 0 and $tokens[0] eq 'END') {
            # This should be an 'END' clause.
            return;
        }
        elsif ($#tokens == 1 && ($tokens[0] eq 'DEFAULT' || $tokens[0] eq 'INCLUDE')) {
            # This is a INCLUDE/DEFAULT clause.
            if (! defined $self->{$name}{'DEFAULT'}) {
                $self->{$name}{'DEFAULT'} = $tokens[1];
            } else {
                $self->{$name}{'DEFAULT'} = $self->{$name}{'DEFAULT'} . " $tokens[1]";
            }
        }
        elsif ($#tokens >= 1) {
            my $property = shift @tokens;
            if ($tokens[0] eq '@{') {
                my @prop = &read_property_block(\*CONFIG, \%locals);
                $self->{$name}{$property} = "@prop";
            } else {
                $self->{$name}{$property} = "@tokens";
            }

            # Shove the property just found into the local macro table.
            $self->define($property,  $self->{$name}{$property}, \%locals);
        } else {
            # Single lined entry.
            die "Malformed configuration file near line $line_number:\n\t@tokens\n\tExpecting 'END'?";
        }

    }
}

# Handle the various directives available.
sub config_handle_directive {
    my $self = shift;
    my $CONFIG = shift;
    my $block = shift;
    my @tokens = @_;

    my ($name, $property, $test_result);

    if ($tokens[0] eq '@@define') {
        # Assign a variable.
        shift @tokens;
        my $var = "\\\@\\\$$tokens[0]";
        shift @tokens;
        $variables{$var} = "@tokens";

    # Now, The conditionals.
    } elsif ($tokens[0] =~ "^\@\@if") {
        if ($tokens[0] eq '@@ifdefined') {
            if (defined $variables{"\\\@\\\$$tokens[1]"}) {
                $test_result = 1;
            } else {
                $test_result = 0;
            }
            shift @tokens; shift @tokens;
        } elsif ($tokens[0] eq '@@ifndefined') {
            if (! defined $variables{"\\\@\\\$$tokens[1]"}) {
                $test_result = 1;
            } else {
                $test_result = 0;
            }
            shift @tokens; shift @tokens;
        } elsif ($tokens[0] eq '@@ifequal') {
            if ($tokens[1] eq $tokens[2]) {
                $test_result = 1;
            } else {
                $test_result = 0;
            }
            shift @tokens; shift @tokens; shift @tokens;
        } elsif ($tokens[0] eq '@@ifnequal') {
            if ($tokens[1] ne $tokens[2]) {
                $test_result = 1;
            } else {
                $test_result = 0;
            }
            shift @tokens; shift @tokens; shift @tokens;
        } else {
            die "Unknown directive line near line $line_number: @tokens\n";
        }

        # Now that we've decided on the truth, let's do something.
        if ($test_result) {
            if ($tokens[0] =~ "^\@\@") {
                # @@ifequal @$HOSTNAME wazor   @@define local_dir /usr/local
                &config_handle_directive($self, $CONFIG, $block, @tokens);
            } elsif ($block =~ "^DEFAULT") {
                $property = shift @tokens;
                if ($tokens[0] eq '@{') {
                    my @prop = &read_property_block(\*CONFIG);
                    $self->{$block}{$property} = "@prop";
                } else {
                    $self->{$block}{$property} = "@tokens";
                }
            } elsif ($block =~ "^NAME") {
                (undef, $name) = split(/ /, $block);
                $property = shift @tokens;
                if ($tokens[0] eq '@{') {
                    my @prop = &read_property_block(\*CONFIG);
                    $self->{$name}{$property} = "@prop";
                } else {
                    $self->{$name}{$property} = "@tokens";
                }
            } else {
                die "Total confusion near line $line_number (missing @@-directive?)\n\tYou cannot assign a property outside a block.\n\tbad line contains: @tokens\n";
            }
        } else {
            # Well, the test was false, but we still need to make sure
            # that there are no propery_blocks involved here.  If
            # there are, they need to be munched up (and ignored).
            if ($tokens[$#tokens] eq '@{') {
                # Munch.
                my @tmp = &read_property_block(\*CONFIG);
            }
        }
    }
    else {
        die "Unknown directive near line $line_number: @tokens\n";
    }
}


# Read until the delimiting @}
sub read_property_block {
    my ($CONFIG, $locals) = @_;
    my (@tmp, $line);

    $line = <CONFIG>;
    while ($line !~ '@}' && ! eof $CONFIG) {
        # Substitute @$-variables in here.
        my $key;

        # First, do substitutions on locals, if there are any.
        if (defined $locals) {
            foreach $key (keys %{$locals}) {
                $line =~ s/$key/$locals->{$key}/g;
            }
        }

        foreach $key (keys %variables) {
            $line =~ s/$key/$variables{$key}/g;
        }

        $tmp[$#tmp + 1] = $line;
        $line_number++;
        $line = <CONFIG>;
    }

    return @tmp;
}


# Basic.  Not too smart right now.
sub define_default_variables {
    my $tmp;

    $variables{"\\\@\\\$COMMENT"} = '#';
    $variables{"\\\@\\\$DATE"} = scalar localtime time;
    $tmp = `/bin/date +%y%b%d 2>/dev/null`;
    chop $tmp;
    $variables{"\\\@\\\$DATETAG"} = $tmp;
    
    # Assign some default macros if uname exists.
    if (-x '/bin/uname') {
        $tmp = `/bin/uname`;
        chop $tmp;
        $variables{"\\\@\\\$OS"}   = $tmp;
        
        $tmp = `/bin/uname -n`;
        chop $tmp;
        $variables{"\\\@\\\$HOSTNAME"}   = $tmp;
        
        $tmp = `/bin/uname -r`;
        chop $tmp;
        $variables{"\\\@\\\$OS_VERSION"} = $tmp;
        $variables{"\\\@\\\$OSVERSION"} = $variables{"\\\@\\\$OS"}
        . $variables{"\\\@\\\$OS_VERSION"};
        
        $tmp = `/bin/uname -p 2>/dev/null`;
        chop $tmp;
        $variables{"\\\@\\\$ARCH"} = $tmp;
    
        # Create these little abbreviations for OSes I like.  If it isn't
        # here, @$OS_SHORT is the same as @$OSVERSION.
        
        if ($variables{"\\\@\\\$OS"} eq "SunOS") {
            if ($variables{"\\\@\\\$OS_VERSION"} =~ /^5.*/) {
                $variables{"\\\@\\\$OS_SHORT"} = 'solaris';
            } elsif ($variables{"\\\@\\\$OS_VERSION"} =~ /^4.*/) {
                $variables{"\\\@\\\$OS_SHORT"} = 'sunos';
            } else {
                $variables{"\\\@\\\$OS_SHORT"} = $variables{"\\\@\\\$OSVERSION"};
            }
        } else {
            $variables{"\\\@\\\$OS_SHORT"} = $variables{"\\\@\\\$OSVERSION"};
        }
    }
}



1;
